/*
   ILI9341 LCD routines for STM32F4xx-nucleo
*/
#include "xstm32f4xx.h"
#include "spi.h"

#define DEF_FONT_SIZE  16

void  delay( int  msec );
//void  name_syslog_posi( int x0, int y0, int x1, int y1 );


void  lcd_dev_init( void )
{
  /* set clock enable */
  RCC->AHB1ENR |= (  RCC_AHB1ENR_GPIOAEN
                   | RCC_AHB1ENR_GPIOBEN
                  ) ;

  /** for LCD_RS:PB3 **/
  GPIOB->MODER   &= ~(0x3 << (3*2));
  GPIOB->MODER   |=  (0x1 << (3*2));
  GPIOB->OTYPER  &= ~(0x1 << 3);
  GPIOB->OTYPER  |=  (0x0 << 3);
  GPIOB->OSPEEDR &= ~(0x3 << (3*2));
  GPIOB->OSPEEDR |=  (0x2 << (3*2));
  GPIOB->PUPDR   &= ~(0x3 << (3*2));
  GPIOB->PUPDR   |=  (0x1 << (3*2));
  GPIOB->BSRR = (0x1 << 3) ;       /* RS:1 */

  /** for LCD_RST:PA10 **/
  GPIOA->MODER   &= ~(0x3 << (10*2));
  GPIOA->MODER   |=  (0x1 << (10*2));
  GPIOA->OTYPER  &= ~(0x1 << 10);
  GPIOA->OTYPER  |=  (0x0 << 10);
  GPIOA->OSPEEDR &= ~(0x3 << (10*2));
  GPIOA->OSPEEDR |=  (0x2 << (10*2));
  GPIOA->PUPDR   &= ~(0x3 << (10*2));
  GPIOA->PUPDR   |=  (0x1 << (10*2));
  GPIOA->BSRR = (0x1 << 10) ;       /* RST:1 */
  GPIOA->BSRR = (0x1 << 10) <<16 ;  /* RST:0 */
  GPIOA->BSRR = (0x1 << 10) ;       /* RST:1 */
}


void  xlcd_set_rs( int flag )
{
  if( flag != 0 ) {
    GPIOB->BSRR = (0x1 << 3) ;       /* RS:1 */
  } else {
    GPIOB->BSRR = (0x1 << 3) <<16 ;  /* RS:0 */
  }
}


void  sendCMD8( unsigned char  com )
{
  xspi_set_cs_lcd( 0 ); /* CS:0 */

  xlcd_set_rs( 0 ); /* RS:0 */

  xspi_send( (uint16_t)com );

  xspi_set_cs_lcd( 1 ); /* CS:1 */
}


void  sendData8( unsigned char  dat )
{
  xspi_set_cs_lcd( 0 ); /* CS:0 */

  xlcd_set_rs( 1 ); /* RS:1 */

  xspi_send( (uint16_t)dat );

  xspi_set_cs_lcd( 1 ); /* CS:1 */
}


void  lcd_setblock( unsigned short int x0, unsigned short int x1, unsigned short int y0, unsigned short int y1 )
{
  //name_syslog_posi( x0, y0, x1, y1 );

  sendCMD8(0x2a);
  sendData8(x0>>8);
  sendData8(x0);
  sendData8(x1>>8);
  sendData8(x1);

  sendCMD8(0x2b);
  sendData8(y0>>8);
  sendData8(y0);
  sendData8(y1>>8);
  sendData8(y1);

  sendCMD8(0x2c);
}


/*
 * color : RGB565
 */
void  lcd_pset( int x, int y, int color )
{
  lcd_setblock( x, x, y, y );

  xspi_set_cs_lcd( 0 );  /* CS:0 */
  xlcd_set_rs( 1 ) ; /* RS:1 */

  xspi_send( (color>>8 ) & 0xff );
  xspi_send( (color    ) & 0xff );

  xspi_set_cs_lcd( 1 );  /* CS:1 */
}


static  unsigned char  dat_bit[8] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 } ;

#include  "fnt_shnm7x14a.inc"


/*
 *  fg : RGB565
 *  bg : RGB565
 */
void  lcd_pset_fontaR( unsigned char *font, int x, int y, int fg, int bg )
{
  uint16_t  fg0, fg1;
  uint16_t  bg0, bg1;
  int  i, k;

  fg0 = (fg>>8) & 0xff;
  fg1 = (fg   ) & 0xff;
  bg0 = (bg>>8) & 0xff;
  bg1 = (bg   ) & 0xff;

  lcd_setblock( x, x+14-1, y-8+1, y );

  xspi_set_cs_lcd( 0 );  /* CS:0 */
  xlcd_set_rs( 1 ) ; /* RS:1 */

  for( k=7 ; k>=0 ; k-- ) {               /* bit pos */
    for( i=0 ; i<14 ; i++ ) {  /* line */
      if( (font[i] & dat_bit[k]) != 0 ) {
        xspi_send( fg0 );
        xspi_send( fg1 );
      } else {
        xspi_send( bg0 );
        xspi_send( bg1 );
      }
    }
  }

  xspi_set_cs_lcd( 1 );  /* CS:1 */
}


void  lcd_put_stringR( int x, int y, int fg, int bg, unsigned char *msg, int size )
{
  int  pos ;
  unsigned char  *fontdata ;
  unsigned int  code ;

  pos = 0 ;
  while( pos < size ) {
    code = msg[pos++] ;
    if( code == 0x0d || code == 0x0a ) {
      code = ' ' ;
    }
    fontdata = fnt_shinm7x14a + (code * 14);
    lcd_pset_fontaR( fontdata, x, y, fg, bg );
    y -= DEF_FONT_SIZE/2;
  }
}


/*
 *  fg : RGB565
 *  bg : RGB565
 */
void  lcd_pset_fonta( unsigned char *font, int x, int y, int fg, int bg )
{
  uint16_t  fg0, fg1;
  uint16_t  bg0, bg1;
  int  i, k;

  fg0 = (fg>>8) & 0xff;
  fg1 = (fg   ) & 0xff;
  bg0 = (bg>>8) & 0xff;
  bg1 = (bg   ) & 0xff;

  lcd_setblock( x, x+8-1, y-14+1, y );

  xspi_set_cs_lcd( 0 );  /* CS:0 */
  xlcd_set_rs( 1 ) ; /* RS:1 */

  for( i=0 ; i<14 ; i++ ) {  /* line */
    for( k=0 ; k<=7 ; k++ ) {               /* bit pos */
      if( (font[i] & dat_bit[k]) != 0 ) {
        xspi_send( fg0 );
        xspi_send( fg1 );
      } else {
        xspi_send( bg0 );
        xspi_send( bg1 );
      }
    }
  }

  xspi_set_cs_lcd( 1 );  /* CS:1 */
}


void  lcd_put_string( int x, int y, int fg, int bg, unsigned char *msg, int size )
{
  int  pos ;
  unsigned char  *fontdata ;
  unsigned int  code ;

  pos = 0 ;
  while( pos < size ) {
    code = msg[pos++] ;
    if( code == 0x0d || code == 0x0a ) {
      code = ' ' ;
    }
    fontdata = fnt_shinm7x14a + (code * 14);
    lcd_pset_fonta( fontdata, x, y, fg, bg );
    x += DEF_FONT_SIZE/2;
  }
}


/*
 *  fg : RGB565
 *  bg : RGB565
 */
void  lcd_pset_fonta2( unsigned char *font, int x, int y, int fg, int bg )
{
  uint16_t  fg0, fg1;
  uint16_t  bg0, bg1;
  int  i, k;

  fg0 = (fg>>8) & 0xff;
  fg1 = (fg   ) & 0xff;
  bg0 = (bg>>8) & 0xff;
  bg1 = (bg   ) & 0xff;

  lcd_setblock( x, x+(8*2)-1, y, y+(14*2)-1 );

  xspi_set_cs_lcd( 0 );  /* CS:0 */
  xlcd_set_rs( 1 ) ; /* RS:1 */

  for( i=0 ; i<14 ; i++ ) {  /* line */
    for( k=0 ; k<=7 ; k++ ) {               /* bit pos */
      if( (font[i] & dat_bit[k]) != 0 ) {
        xspi_send( fg0 );
        xspi_send( fg1 );
        xspi_send( fg0 );
        xspi_send( fg1 );
      } else {
        xspi_send( bg0 );
        xspi_send( bg1 );
        xspi_send( bg0 );
        xspi_send( bg1 );
      }
    }
    for( k=0 ; k<=7 ; k++ ) {               /* bit pos */
      if( (font[i] & dat_bit[k]) != 0 ) {
        xspi_send( fg0 );
        xspi_send( fg1 );
        xspi_send( fg0 );
        xspi_send( fg1 );
      } else {
        xspi_send( bg0 );
        xspi_send( bg1 );
        xspi_send( bg0 );
        xspi_send( bg1 );
      }
    }
  }

  xspi_set_cs_lcd( 1 );  /* CS:1 */
}


void  lcd_put_string2( int x, int y, int fg, int bg, unsigned char *msg, int size )
{
  int  pos ;
  unsigned char  *fontdata ;
  unsigned int  code ;

  pos = 0 ;
  while( pos < size ) {
    code = msg[pos++] ;
    if( code == 0x0d || code == 0x0a ) {
      code = ' ' ;
    }
    fontdata = fnt_shinm7x14a + (code * 14);
    lcd_pset_fonta2( fontdata, x, y, fg, bg );
    x += DEF_FONT_SIZE/2 * 2;
  }
}


void  lcd_dispbuff( unsigned char *buff )
{
  int  i;

  lcd_setblock( 0, 240-1, 0, 260 );

  xspi_set_cs_lcd( 0 );  /* CS:0 */
  xlcd_set_rs( 1 ) ; /* RS:1 */

  for( i=0 ; i<(240*261) ; i++ ) {
    if( *buff == 0 ) {
      xspi_send( (uint16_t)0x00 );
      xspi_send( (uint16_t)0x00 );
    } else {
      xspi_send( (uint16_t)0xff );
      xspi_send( (uint16_t)0xff );
    }

    buff++;
  }

  xspi_set_cs_lcd( 1 );  /* CS:1 */
}


void  lcd_fill( int x0, int x1, int y0, int y1, int color )
{
  uint16_t  fg0, fg1;
  int  i;

  fg0 = (color>>8) & 0xff;
  fg1 = (color   ) & 0xff;

  lcd_setblock( x0, x1, y0, y1 );

  xspi_set_cs_lcd( 0 );  /* CS:0 */
  xlcd_set_rs( 1 ) ; /* RS:1 */

  for( i=0 ; i<((x1-x0+1)*(y1-y0+1)) ; i++ ) {
    xspi_send( fg0 );
    xspi_send( fg1 );
  }

  xspi_set_cs_lcd( 1 );  /* CS:1 */
}


/*
 * color : RGB565
 */
void  lcd_fill_full( int color )
{
  uint16_t  fg0, fg1;
  int  i;

  fg0 = (color>>8) & 0xff;
  fg1 = (color   ) & 0xff;

  lcd_setblock( 0, 240-1, 0, 320-1 );

  xspi_set_cs_lcd( 0 );  /* CS:0 */
  xlcd_set_rs( 1 ) ; /* RS:1 */

  for( i=0 ; i<240*320 ; i++ ) {
    xspi_send( fg0 );
    xspi_send( fg1 );
  }

  xspi_set_cs_lcd( 1 );  /* CS:1 */
}


/*
 * RGB565 BMP  240x320
 */
void  lcd_set_rgb565_bmp( unsigned char  *bmp )
{
  int  i;
  unsigned char  *pos;

  pos = bmp + 66 + (320 * 240) * 2 - 1;

  lcd_setblock( 0, 240-1, 0, 320-1 );

  xspi_set_cs_lcd( 0 );  /* CS:0 */
  xlcd_set_rs( 1 ) ; /* RS:1 */

  for( i=0 ; i<240*320 ; i++ ) {
    xspi_send( *pos-- );
    xspi_send( *pos-- );
  }

  xspi_set_cs_lcd( 1 );  /* CS:1 */
}


/*
 * RGB565 BMP  240x240
 */
void  lcd_set_rgb565_bmp240x240( unsigned char  *bmp )
{
  int  i;
  unsigned char  *pos;

  pos = bmp + 66 + (240 * 240) * 2 - 1;

  lcd_setblock( 0, 240-1, 0, 240-1 );

  xspi_set_cs_lcd( 0 );  /* CS:0 */
  xlcd_set_rs( 1 ) ; /* RS:1 */

  for( i=0 ; i<240*240 ; i++ ) {
    xspi_send( *pos-- );
    xspi_send( *pos-- );
  }

  xspi_set_cs_lcd( 1 );  /* CS:1 */
}


/** for ili9341 **/
void  lcd_init( void )
{
  lcd_dev_init();

  sendCMD8(0x01); // Engage software reset
  delay(150);

  sendCMD8(0xEF);
  sendData8(0x03);
  sendData8(0x80);
  sendData8(0x02);

  sendCMD8(0xCB);
  sendData8(0x39);
  sendData8(0x2C);
  sendData8(0x00);
  sendData8(0x34);
  sendData8(0x02);

  sendCMD8(0xCF);
  sendData8(0x00);
  sendData8(0XC1);
  sendData8(0X30);

  sendCMD8(0xE8);
  sendData8(0x85);
  sendData8(0x00);
  sendData8(0x78);

  sendCMD8(0xEA);
  sendData8(0x00);
  sendData8(0x00);
 
  sendCMD8(0xED);
  sendData8(0x64);
  sendData8(0x03);
  sendData8(0X12);
  sendData8(0X81);

  sendCMD8(0xF7);
  sendData8(0x20);
  
  sendCMD8(0xC0);    //Power control 
  sendData8(0x23);   //VRH[5:0] 
 
  sendCMD8(0xC1);    //Power control 
  sendData8(0x10);   //SAP[2:0];BT[3:0] 

  sendCMD8(0xC5);    //VCM control 
  sendData8(0x3e);   //Contrast
  sendData8(0x28); 
 
  sendCMD8(0xC7);    //VCM control2 
  sendData8(0x86);   //--
 
  sendCMD8(0x36);    // Memory Access Control 
  sendData8(0x48);

  sendCMD8(0x37);
  sendData8(0x00);
  
  sendCMD8(0x3A);
  sendData8(0x55);

  sendCMD8(0xB1);
  sendData8(0x00);
  sendData8(0x18);
 
  sendCMD8(0xB6);    // Display Function Control 
  sendData8(0x08);
  sendData8(0x82);
  sendData8(0x27);

  sendCMD8(0x11);    //Exit Sleep 
  delay(120); 

  sendCMD8(0x29);    //Display on 

  sendCMD8(0x2c); 

  /* clear LCD */
  lcd_fill_full( 0x0000 );
}
